#ifndef TIMING_H_
#define TIMING_H_

/**
 * check_timing(void):
 * Check that pthread_create_blocking_np() results in the expected relative
 * timing.
 */
int check_timing(void);

#endif /* !TIMING_H_ */
